==> mindstudio_probe-8.1.0-py3-none-any.whl: msprobe/docs/33.generate_operator_MindSpore.md <==
# 单算子API自动生成脚本

## 1 简介

单算子API自动生成脚本通过提取dump数据中的可疑算子，对其进行单API复现，输出单API精度的比对结果。具体而言，该工具可以从dump数据中提取可疑API的前反向信息，根据前反向数据生成单API的前反向过程，最后通过**新精度标准比对法**<sup>a</sup>将 NPU 和 CPU 的结果进行比对，从而给出不同比对方法下的比对结果。本工具支持**随机生成模式和真实数据模式**<sup>b</sup>。

a. 在生成单API脚本时可以选择由工具构造随机数获得 dump 数据或选择真实输入的数据进行单API复现。随机生成模式（对应 task: "statistics"）执行效率高，可以快速获得结果，但数据精度低，只能大致判断精度问题；真实数据模式（对应 task: "tensor"）执行效率略低于随机生成模式，但是数据精度高，可以准确判断精度问题。

## 2 使用方式

### 前提
1. 安装 msprobe。详见[ msprobe 安装](./01.installation.md)章节。
2. 已完成对训练过程的dump，获得dump.json文件。
    [MindSpore 场景下的数据采集](./06.data_dump_MindSpore.md)章节或[Msadapter 场景下的数据采集](./29.data_dump_MSAdapter.md)章节，注意需要配置 level="L1"。

3. 发现某个算子疑似存在精度问题，并得知算子名，如Mint.split.1、Functional.softmax.3、Tensor.add.0、Torch.matmul.5等,要求API入参与torch场景相同。

4.（可选）当需要使用Msadapter时，由于需要环境中同时存在 Torch 与 Msadapter，所以只支持在**安装原生Torch**的场景下通过export PYTHONPATH="xx/msadapter/build/lib"等通过**环境变量使能Msadapter的方式**的环境中进行预检,预检工具能够自动索引得到所需的 Torch 与 Msadapter环境,环境安装详细参考:[msadapter官网](https://gitee.com/mindspore/msadapter)。

### 2.1 dump.json 示例（由dump生成）
```
{
    "task": "statistics",
    "level": "L1",
    "framework": "mindtorch",
    "dump_data_dir": null,
    "data": {
        "Tensor.reshape.779.forward": {
            "input_args": [
             {
              "type": "mindspore.Tensor",
              "dtype": "BFloat16",
              "shape": [
               8192,
               896
              ],
              "Max": 0.62890625,
              "Min": -0.78515625,
              "Mean": 0.00035858154296875,
              "Norm": 105.0
             },
             {
              "type": "int",
              "value": -1
             },
             {
              "type": "int",
              "value": 896
             }
            ],
            "input_kwargs": {},
            "output": [
             {
              "type": "mindspore.Tensor",
              "dtype": "BFloat16",
              "shape": [
               8192,
               896
              ],
              "Max": 0.62890625,
              "Min": -0.78515625,
              "Mean": 0.00035858154296875,
              "Norm": 105.0
             }
            ]
           },
           "Tensor.reshape.779.backward": {
            "input": [
             {
              "type": "mindspore.Tensor",
              "dtype": "BFloat16",
              "shape": [
               8192,
               896
              ],
              "Max": 0.0,
              "Min": 0.0,
              "Mean": 0.0,
              "Norm": 0.0
             }
            ],
            "output": [
             {
              "type": "mindspore.Tensor",
              "dtype": "BFloat16",
              "shape": [
               8192,
               896
              ],
              "Max": 0.0,
              "Min": 0.0,
              "Mean": 0.0,
              "Norm": 0.0
             }
            ]
           }
        }
    }
```

### 2.2 配置config_op.json
单API复现参数配置如下（以复现softmax算子为例）：
```
{
    "dump_json_path": "./dump.json",
    "api_name": "Tensor.reshape.779",
    "extract_api_path": "Tensor.reshape.779",
    "propagation": "backward",
    "data_mode": "random_data",
    "random_seed": 42,
    "iter_times": 1
}
```
**配置文件参数说明**

   | 参数名称                     | 解释                                                                                                                | 
   | ---------------------------- |-------------------------------------------------------------------------------------------------------------------| 
   | dump_json_path   | dump.json的文件路径，包含所有dump算子的信息；如果已经提取了可疑算子并保存可以不指定。                                                                 |
   | api_name             | 算子名（目前只支持Mint，Tensor，Functional，Torch类中可自动求导api），如Mint.split.1，Functional.softmax.3、Tensor.add.0、Torch.matmul.5等。 | 
   | extract_api_path               | 提取可疑算子的json文件路径                                                                                                   | 
   | propagation | 选择复现算子的forward还是backward，默认为forward                                                                               | 
   | data_mode                 | 选择复现算子的随机数据（random_data）还是真实数据（real_data）模式，默认为random_data                                                        |
   | random_seed | 仅random_data模式有效，表示手动设定的随机种子，默认为1234                                                                              | 
   | iter_times             | 仅random_data模式有效，表示单API运行的次数，由于安全相关原因，最大支持设置为1000                                                                 |

 ### 2.3 运行命令生成单API脚本
config_op.json配置好后，运行如下命令：
```
msprobe -f mindspore op_generate -i ./config.json -o ./
```
或者

进入到mstt的generate_op_script文件夹
```
cd mstt/debug/accuracy_tools/msprobe/mindspore/api_accuracy_checker/generate_op_script
```
运行
```
python op_generator.py -i ./config_op.json -o ./
```
**参数说明**
   | 参数名称                     | 解释                                                         | 是否必选                           |
   | ---------------------------- | ------------------------------------------------------------ | ---------------------------------- |
   | -i 或 --config_input   | config_op.json的路径                                   | 是                                 |
   | -o 或 --api_output_path             | 单API脚本的输出路径                            | 是                                 |
 
### 2.4 运行单API脚本
 运行完op_generator.py后，会在指定路径下生成api_name.py的单API脚本，例如Mint.split.1.forward.py、Functional.softmax.3.backward.py、Tensor.add.0.forward.py、Torch.matmul.5.backward.py

运行单API脚本即可获得不同比对方法下的比对结果
```
python api_name.py
```

**运行结果说明**

单算子脚本生成到路径`./op_result_output`的 `accuracy_checking_result_{timestamp}.csv` 和 `accuracy_checking_details_{timestamp}.csv` 文件内容详情如下：

`accuracy_checking_details_{timestamp}.csv`

| 字段      | 含义       |
| ------------------- | ------------------------------------------------------------ |
| API Name            | API 名称。                                        |
| Bench Dtype         | 标杆数据的 API 数据类型。      |
| Tested Dtype        | 被检验数据的 API 数据类型。                                  |
| Shape               | API 的 Shape 信息。       |
| Cosine              | 被检验数据与标杆数据的余弦相似度。                         |
| MaxAbsErr           | 被检验数据与标杆数据的最大绝对误差。                       |
| MaxRelativeErr      | 被检验数据与标杆数据的最大相对误差。                     |
| Status              | API 预检通过状态，pass 表示通过测试，error 表示未通过。 |
| Message             | 提示信息。            |

注意：PyTorch 无法对 dtype 为整数类型的 tensor 进行反向求导，而 MindSpore 支持。反向过程的预检仅比较 dtype 为浮点型的输出。

`accuracy_checking_result_{timestamp}.csv`

| 字段                  | 含义      |
| --------------------- | ----------------- |
| API Name              | API 名称。         |
| Forward Test Success  | 前向 API 是否通过测试，pass 为通过，error 为错误。 |
| Backward Test Success | 反向 API 是否通过测试，pass 为通过，error 为错误，如果是空白的话代表该 API 没有反向输出。 |
| Message               | 提示信息。         |

Forward Test Success 和 Backward Test Success 是否通过测试是由 `accuracy_checking_details_{timestamp}.csv` 中的余弦相似度、最大绝对误差判定结果决定的。具体规则详见 [4.1 API 预检指标](#41-api-预检指标)。
需要注意的是 `accuracy_checking_details_{timestamp}.csv` 中可能存在一个 API 的前向（反向）有多个输出，那么每个输出记录一行，而在 `accuracy_checking_result_{timestamp}.csv` 中的结果需要该 API 的所有结果均为 pass 才能标记为 pass，只要存在一个 error 则标记 error。

### 3.1 API 预检指标

   - API 预检指标是通过对 `accuracy_checking_details_{timestamp}.csv` 中的余弦相似度、最大绝对误差的数值进行判断，得出该 API 是否符合精度标准的参考指标。
   - 余弦相似度大于 0.99，并且最大绝对误差小于 0.0001，标记“pass”，否则标记为“error”。

==> mindstudio_probe-8.1.1-py3-none-any.whl: msprobe/docs/33.generate_operator_MindSpore.md <==
# 单算子API自动生成脚本

## 1 简介

单算子API自动生成脚本通过提取dump数据中的可疑算子，对其进行单API复现，输出单API精度的比对结果。具体而言，该工具可以从dump数据中提取可疑API的前反向信息，根据前反向数据生成单API的前反向过程，最后通过**新精度标准比对法**<sup>a</sup>将 NPU 和 CPU 的结果进行比对，从而给出不同比对方法下的比对结果。本工具支持**随机生成模式和真实数据模式**<sup>b</sup>。

a. 在生成单API脚本时可以选择由工具构造随机数获得 dump 数据或选择真实输入的数据进行单API复现。随机生成模式（对应 task: "statistics"）执行效率高，可以快速获得结果，但数据精度低，只能大致判断精度问题；真实数据模式（对应 task: "tensor"）执行效率略低于随机生成模式，但是数据精度高，可以准确判断精度问题。

## 2 使用方式

### 前提
1. 安装 msprobe。详见[ msprobe 安装](./01.installation.md)章节。
2. 已完成对训练过程的dump，获得dump.json文件。
    [MindSpore 场景下的数据采集](./06.data_dump_MindSpore.md)章节或[Msadapter 场景下的数据采集](./29.data_dump_MSAdapter.md)章节，注意需要配置 level="L1"。

3. 发现某个算子疑似存在精度问题，并得知算子名，如Mint.split.1、Functional.softmax.3、Tensor.add.0、Torch.matmul.5等,要求API入参与torch场景相同。

4.（可选）当需要使用Msadapter时，由于需要环境中同时存在 Torch 与 Msadapter，所以只支持在**安装原生Torch**的场景下通过export PYTHONPATH="xx/msadapter/build/lib"等通过**环境变量使能Msadapter的方式**的环境中进行预检,预检工具能够自动索引得到所需的 Torch 与 Msadapter环境,环境安装详细参考:[msadapter官网](https://gitee.com/mindspore/msadapter)。

### 2.1 dump.json 示例（由dump生成）
```
{
    "task": "statistics",
    "level": "L1",
    "framework": "mindtorch",
    "dump_data_dir": null,
    "data": {
        "Tensor.reshape.779.forward": {
            "input_args": [
             {
              "type": "mindspore.Tensor",
              "dtype": "BFloat16",
              "shape": [
               8192,
               896
              ],
              "Max": 0.62890625,
              "Min": -0.78515625,
              "Mean": 0.00035858154296875,
              "Norm": 105.0
             },
             {
              "type": "int",
              "value": -1
             },
             {
              "type": "int",
              "value": 896
             }
            ],
            "input_kwargs": {},
            "output": [
             {
              "type": "mindspore.Tensor",
              "dtype": "BFloat16",
              "shape": [
               8192,
               896
              ],
              "Max": 0.62890625,
              "Min": -0.78515625,
              "Mean": 0.00035858154296875,
              "Norm": 105.0
             }
            ]
           },
           "Tensor.reshape.779.backward": {
            "input": [
             {
              "type": "mindspore.Tensor",
              "dtype": "BFloat16",
              "shape": [
               8192,
               896
              ],
              "Max": 0.0,
              "Min": 0.0,
              "Mean": 0.0,
              "Norm": 0.0
             }
            ],
            "output": [
             {
              "type": "mindspore.Tensor",
              "dtype": "BFloat16",
              "shape": [
               8192,
               896
              ],
              "Max": 0.0,
              "Min": 0.0,
              "Mean": 0.0,
              "Norm": 0.0
             }
            ]
           }
        }
    }
```

### 2.2 配置config_op.json
单API复现参数配置如下（以复现softmax算子为例）：
```
{
    "dump_json_path": "./dump.json",
    "api_name": "Tensor.reshape.779",
    "extract_api_path": "Tensor.reshape.779",
    "propagation": "backward",
    "data_mode": "random_data",
    "random_seed": 42,
    "iter_times": 1
}
```
**配置文件参数说明**

   | 参数名称                     | 解释                                                                                                                                                    | 
   | ---------------------------- |-------------------------------------------------------------------------------------------------------------------------------------------------------| 
   | dump_json_path   | dump.json的文件路径，包含所有dump算子的信息；如果已经提取了可疑算子并保存可以不指定。                                                                                                     |
   | api_name             | 算子名（目前MindSpore支持类型包括：Mint，Tensor，Msadapter支持类型包括：Tensor，Functional，Torch类中可自动求导api），如Mint.split.1，Functional.softmax.3、Tensor.add.0、Torch.matmul.5等。 | 
   | extract_api_path               | 提取可疑算子的json文件路径                                                                                                                                       | 
   | propagation | 选择复现算子的forward还是backward，默认为forward                                                                                                                   | 
   | data_mode                 | 选择复现算子的随机数据（random_data）还是真实数据（real_data）模式，默认为random_data                                                                                            |
   | random_seed | 仅random_data模式有效，表示手动设定的随机种子，默认为42                                                                                                                    | 
   | iter_times             | 仅random_data模式有效，表示单API运行的次数，由于安全相关原因，最大支持设置为1000                                                                                                     |

 ### 2.3 运行命令生成单API脚本
config_op.json配置好后，运行如下命令：
```
msprobe -f mindspore op_generate -i ./config_op.json -o ./
```

**参数说明**
   | 参数名称                     | 解释                                                         | 是否必选                           |
   | ---------------------------- | ------------------------------------------------------------ | ---------------------------------- |
   | -i 或 --config_input   | config_op.json的路径                                   | 是                                 |
   | -o 或 --api_output_path             | 单API脚本的输出路径                            | 是                                 |
 
### 2.4 运行单API脚本
 运行完op_generator.py后，会在指定路径下生成api_name.py的单API脚本，例如Mint.split.1.forward.py、Functional.softmax.3.backward.py、Tensor.add.0.forward.py、Torch.matmul.5.backward.py

运行单API脚本即可获得不同比对方法下的比对结果
```
python api_name.py
```

**运行结果说明**

单算子脚本生成到路径`./op_result_output`的 `accuracy_checking_result_{timestamp}.csv` 和 `accuracy_checking_details_{timestamp}.csv` 文件内容详情如下：

`accuracy_checking_details_{timestamp}.csv`

| 字段      | 含义       |
| ------------------- | ------------------------------------------------------------ |
| API Name            | API 名称。                                        |
| Bench Dtype         | 标杆数据的 API 数据类型。      |
| Tested Dtype        | 被检验数据的 API 数据类型。                                  |
| Shape               | API 的 Shape 信息。       |
| Cosine              | 被检验数据与标杆数据的余弦相似度。                         |
| MaxAbsErr           | 被检验数据与标杆数据的最大绝对误差。                       |
| MaxRelativeErr      | 被检验数据与标杆数据的最大相对误差。                     |
| Status              | API 预检通过状态，pass 表示通过测试，error 表示未通过。 |
| Message             | 提示信息。            |

注意：PyTorch 无法对 dtype 为整数类型的 tensor 进行反向求导，而 MindSpore 支持。反向过程的预检仅比较 dtype 为浮点型的输出。

`accuracy_checking_result_{timestamp}.csv`

| 字段                  | 含义      |
| --------------------- | ----------------- |
| API Name              | API 名称。         |
| Forward Test Success  | 前向 API 是否通过测试，pass 为通过，error 为错误。 |
| Backward Test Success | 反向 API 是否通过测试，pass 为通过，error 为错误，如果是空白的话代表该 API 没有反向输出。 |
| Message               | 提示信息。         |

Forward Test Success 和 Backward Test Success 是否通过测试是由 `accuracy_checking_details_{timestamp}.csv` 中的余弦相似度、最大绝对误差判定结果决定的。具体规则详见 [4.1 API 预检指标](#41-api-预检指标)。
需要注意的是 `accuracy_checking_details_{timestamp}.csv` 中可能存在一个 API 的前向（反向）有多个输出，那么每个输出记录一行，而在 `accuracy_checking_result_{timestamp}.csv` 中的结果需要该 API 的所有结果均为 pass 才能标记为 pass，只要存在一个 error 则标记 error。

### 3.1 API 预检指标

   - API 预检指标是通过对 `accuracy_checking_details_{timestamp}.csv` 中的余弦相似度、最大绝对误差的数值进行判断，得出该 API 是否符合精度标准的参考指标。
   - 余弦相似度大于 0.99，并且最大绝对误差小于 0.0001，标记“pass”，否则标记为“error”。
